#!/usr/bin/perl
# MapTileCacher_IMP.perl, version 0.52
# created by MichaelAtUVa
# updated by Cristian Streng based on suggestions by n95_rit, Kozawa, denilsonsa, and others
# modified by Luca Bisti
# based on original MapTileCacher.perl version 1.37 (http://www.mgmaps.com/cache/) modified tv-arnd

#Externals
use Math::Trig;
use POSIX;

#Proxy Variables
$use_proxy = 0;
# uncomment to use proxy for Google Satellite Download
# $use_proxy = 1;
@proxy_list = (
   'http://edwarf.com/cgi-bin/nph-proxy.cgi/000000A/',
   'http://hostingdi.com/cgiproxy/nph-proxy.pl/000000A/',
   'http://www.alexmattoni.info/scgi-bin/nph-proxy.cgi/000000A/'
);

#Caching Variables
$dir_TopLevel = "MGMapsCache";
$file_CacheConf = "cache.conf";
$cache_Version = 3;
$tiles_per_file = 1;
$hash_size = 1;
$center_found = 0;

#Cache Source Variables
$cache_source = 0;
$dir_TopLevel_src = "MGMapsCache_src";
$file_CacheConf_src = "cache.conf";
$cache_Version_src = 3;
$tiles_per_file_src = 1;
$hash_size_src = 1;
$center_found_src = 0;

# Sleep
$sleepTime = 10;
$sleepEvery = 250;

print <<EOF;

*** MGMaps Tile Cacher ***

EOF

#OS Specific Variables
$pid = $$;
$wget = "wget -Y on -nv -O TempTile.$pid ";
$linuxDirDivider = "\/"; $linuxMoveCommand = "mv"; $linuxDeleteCommand = "rm"; $linuxCopyCommand = "cp";
$windowsDirDivider = "\\"; $windowsMoveCommand = "move"; $windowsDeleteCommand = "del"; $windowsCopyCommand = "copy";
print "Detected OS: $^O\n";

# tv-arnd begin
my $IMmontage;
if($^O eq "darwin") {
  $IMmontage = "/opt/local/bin/montage"; # MacOS X: path and command for ImageMagick montage
}
elsif($^O ge "linux" || $^O ge "cygwin" ){
  $IMmontage ="/usr/bin/montage";        # Linux: path and command for ImageMagick montage
}
# tv-arnd end


if($^O ge "linux" || $^O ge "cygwin" )
{
  $systemDirDivider = $linuxDirDivider;
  $systemMoveCommand = $linuxMoveCommand;
  $systemCopyCommand = $linuxCopyCommand;
  $systemDeleteCommand = $linuxDeleteCommand;
  $systemRefererParam = "--referer=http://maps.google.com/ ";
}
elsif($^O ge "MSWin")
{
  $systemDirDivider = $windowsDirDivider;
  $systemMoveCommand = $windowsMoveCommand;
  $systemCopyCommand = $windowsCopyCommand;
  $systemDeleteCommand = $windowsDeleteCommand;
  $systemRefererParam = "--referer=http://maps.google.com/ ";
}
if($^O eq "darwin")
{
  $systemDirDivider = $linuxDirDivider;
  $systemMoveCommand = $linuxMoveCommand;
  $systemDeleteCommand = $linuxDeleteCommand;
  $wget = "curl -# -o TempTile.$pid ";
  $systemRefererParam = "-e http://maps.google.com/ ";
}

$file_CacheConf_src = $dir_TopLevel_src.$systemDirDivider.$file_CacheConf;
$file_CacheConf = $dir_TopLevel.$systemDirDivider.$file_CacheConf;

# Main


# Check Cache Source configuration
if (-d $dir_TopLevel_src == FALSE) {
  $cache_source = 0;
}
else
{
if (-f $file_CacheConf_src != FALSE) {
  open(F, "<".$file_CacheConf_src) or die("Cannot read cache source config file: $file_CacheConf_src");
  $cache_source = 1;
  $line = <F>;
  $hash_size_src = 97;
  next if ($line =~ /^#/);
  @splits = split(/\s*=\s*/, $line);
  if (($splits[0] eq 'version') && int($splits[1]) > 2) {
    while ($line = <F>) {
      @splits = split('=', $line);
      if ($splits[0] eq 'tiles_per_file') {
        $tiles_per_file_src = int($splits[1]);
      }
      elsif ($splits[0] eq 'hash_size') {
        $hash_size_src = int($splits[1]);
      }
      elsif ($splits[0] eq 'center') {
        $center_found_src = 1;
      }
    }
  }
  else {
    print "Older version of MGMapsCache found for source cache, please (re)move it.\n";
    $cache_source = 0;
    exit;
  }
  close(F);
}
if ($tiles_per_file_src > 1) {
  $hash_size_src = 1;
}
print "Cached source: $tiles_per_file_src tiles per file, hash_size = $hash_size_src\n";
}

# Read and/or write configuration
if (-d $dir_TopLevel == FALSE) {
  system("mkdir $dir_TopLevel");
}
# Search for version number
$configured = 0;
if (-f $file_CacheConf != FALSE) {
  open(F, "<".$file_CacheConf) or die("Cannot read cache config file");
  $line = <F>;
  $hash_size = 97;
  next if ($line =~ /^#/);
  @splits = split(/\s*=\s*/, $line);
  if (($splits[0] eq 'version') && int($splits[1]) > 2) {
    $configured = 1;
    while ($line = <F>) {
      @splits = split('=', $line);
      if ($splits[0] eq 'tiles_per_file') {
        $tiles_per_file = int($splits[1]);
      }
      elsif ($splits[0] eq 'hash_size') {
        $hash_size = int($splits[1]);
      }
      elsif ($splits[0] eq 'center') {
        $center_found = 1;
      }
    }
  }
  else {
    print "Older version of MGMapsCache found, please (re)move it.\n";
    exit;
  }
  close(F);
}

if ($tiles_per_file > 1) {
  $hash_size = 1;
}



# Ask for configuration
if ($configured == 0) {
  # read configuration
  print "Cache configuration file does not exist, configuring...\n\n";
  do {
    print "Map tiles per file (a power of two, 1-4096, default $tiles_per_file): ";
    chomp($tpf = <>);
    if ($tpf eq "") {
      $tpf = $tiles_per_file;
      break;
    }
    $tpf = int($tpf);
    if ($tpf == 0) {
      $tpf = $tiles_per_file;
    }
    # power of two
  } while ($tpf < 1 || $tpf > 4096 || ($tpf & -$tpf) != $tpf);

  $tiles_per_file = $tpf;

  if ($tpf == 1) {
    do {
      print "Hash size (examples: 11, 23, 47, 97 or 1 to disable hashing, default $hash_size): ";
      chomp($hs = <>);
      if ($hs eq "") {
        $hs = $hash_size;
        break;
      }
      $hs = int($hs);
      if ($hs == 0) {
        $hs = $hash_size;
      }
    } while ($hs < 1 || $hs >= 100);

    $hash_size = $hs;
  }

  else {
    print "Hashing is disabled.\n";
    $hash_size = 1;
  }

  open(F, ">".$file_CacheConf) or die("Cannot write cache config file");
  print F "version=".$cache_Version."\n";
  print F "tiles_per_file=".$tiles_per_file."\n";
  print F "hash_size=".$hash_size."\n";
  close(F);
}
else {
  print "Storing $tiles_per_file tiles per file.\n";
  if ($hash_size > 1) {
    print "Hash size is $hash_size.\n";
  }
  else {
    print "Hashing is disabled.\n";
  }
  print "\n";
}

# Cache output computations
$tiles_per_file_log = 0;
$tpf = $tiles_per_file;
while ($tpf > 1) {
  $tiles_per_file_log++;
  $tpf>>=1;
}
$tpfy = $tiles_per_file_log>>1;
$tpfx = 1<<($tpfy + ($tiles_per_file_log&1));
$tpfy = 1<<$tpfy;

# Cache Source computations
$tiles_per_file_log = 0;
$tpf = $tiles_per_file_src;
while ($tpf > 1) {
  $tiles_per_file_log++;
  $tpf>>=1;
}
$tpfy_src = $tiles_per_file_log>>1;
$tpfx_src = 1<<($tpfy_src + ($tiles_per_file_log&1));
$tpfy_src = 1<<$tpfy_src;


my @mapTypeOrder = (
  'GoogleMap', 'GoogleSat', 'GoogleHyb', 'GoogleTer', 'GoogleChina',
  'MicrosoftMap', 'MicrosoftSat', 'MicrosoftHyb', 'MicrosoftTer', 'MicrosoftBrMap',
  'YahooMap', 'YahooSat', 'YahooHyb', 'YahooInMap', 'YahooInHyb',
  'OpenStreetMap', 'Osmarender', 'OpenAerialMap', 'OpenCycleMap'
);

my %mapTypes = (

GoogleMap => 'Google Road Maps',
GoogleSat => 'Google Satellite Images',
GoogleHyb => 'Google Satellite Images with Road Maps Overlayed',
GoogleTer => 'Google Terrain Maps',
GoogleChina => "Google Road Maps (China)\n",

MicrosoftMap => 'Microsoft Live Road Maps',
MicrosoftSat => 'Microsoft Live Aerial Images',
MicrosoftHyb => 'Microsoft Live Aerial Images with Road Maps Overlayed',
MicrosoftTer => 'Microsoft Live Road Maps with Terrain Shading',
MicrosoftBrMap => "Microsoft Live Road Maps (Brasil)\n",

YahooMap => 'Yahoo Road Maps',
YahooSat => 'Yahoo Satellite Images',
YahooHyb => 'Yahoo Satellite Images with Road Maps Overlayed',
YahooInMap => 'Yahoo Road Maps (India)',
YahooInHyb => "Yahoo Satellite Images with Road Maps Overlayed (India)\n",

OpenStreetMap => 'OpenStreetMap.org Maps (Mapnik Renderer)',
Osmarender => "OpenStreetMap.org Maps (Osmarender)",
OpenAerialMap => 'OpenAerialMap.org imagery',
OpenCycleMap => 'Open Cycle Map imagery\n'
);

print <<EOF;
Please choose map type by entering the associated keyword (case sensitive)...

EOF

foreach $k (@mapTypeOrder) {
  print "$k - $mapTypes{$k}\n";
}

print "Choice (default OpenStreetMap): ";
do {
  $MapType = <>;
# tv-arnd begin
 if ($MapType eq "\n") {$MapType = "OpenStreetMap";} # set my default
# tv-arnd end
  $MapType =~ s/^\s+//;
  $MapType =~ s/\s+$//;

  $found = 0;
  while ( my ($k, $v) = each (%mapTypes) ) {
    if (lc($k) eq lc($MapType)) {
      $found = 1;
      $MapType = $k;
      last;
    }
  }

  if ($found == 0) {
    print "Invalid map type. Please try again: ";
  }
} while (($MapType eq "") || ($found == 0));

print "Enter Zoom Level (1-22) (default 15): ";
chomp($zoomLevel = <>);
# tv-arnd begin
  if ($zoomLevel == "") {$zoomLevel = 15;} # set my default
# tv-arnd end
$zoomLevel = int($zoomLevel);

# tv-arnd begin
print "\nHow many Zoom-Levels to process (zoom in) - be careful do not enter too much (default 1): ";
chomp ($ZoomLevels_wanted = <>);
if ($ZoomLevels_wanted == "") {$ZoomLevels_wanted = 1; } # set my default
$ZoomLevels_wanted = int ($ZoomLevels_wanted);
# tv-arnd end

print "\nEnter coordinates in decimal degrees format... 38.031078, -78.481529 \n";

print "Top left (latitude, longitude) (default 49.91, 8.7):  ";
@topLeftCoords = split(/,/,<>,2);
# tv-arnd begin
if (@topLeftCoords == 1) {@topLeftCoords = (49.91, 8.7); }
# tv-arnd end
($topLeftTileX, $topLeftTileY) = convertCoordindates2Tiles(@topLeftCoords, $zoomLevel); # set my default


print "Bottom right (latitude, longitude) (default 49.885, 8.735):  ";
@bottomRightCoords = split(/,/,<>,2);
# tv-arnd begin
if (@bottomRightCoords == 1) {@bottomRightCoords = (49.885, 8.735); };
# tv-arnd end
($bottomRightTileX, $bottomRightTileY) = convertCoordindates2Tiles(@bottomRightCoords, $zoomLevel); # set my default

my $bound = 1<<$zoomLevel;
# swap coords if not in correct order
if ($topLeftTileX > $bottomRightTileX) { $t = $topLeftTileX; $topLeftTileX = $bottomRightTileX; $bottomRightTileX = $t; }
if ($topLeftTileY > $bottomRightTileY) { $t = $topLeftTileY; $topLeftTileY = $bottomRightTileY; $bottomRightTileY = $t; }
# out of map?
if ($topLeftTileX < 0) { $topLeftTileX = 0; }
if ($bottomRightTileX >= $bound) { $bottomRightTileX = $bound-1; }
if ($topLeftTileY < 0) { $topLeftTileY = 0; }
if ($bottomRightTileY >= $bound) { $bottomRightTileY = $bound-1; }
# still swapped coords? (we're entirely out of the map)
if ($topLeftTileX > $bottomRightTileX or $topLeftTileY > $bottomRightTileY) {
  print "\n*** No map tiles to download ***\n";
  exit;
}

$numTiles = (abs($topLeftTileX-$bottomRightTileX)+1)*(abs($topLeftTileY-$bottomRightTileY)+1);
$numTiles *= (index($MapType,"Hyb") != -1 && ($MapType ne "MicrosoftHyb")) ? 2 : 1;

if ($center_found == 0) {
  open(F, ">>".$file_CacheConf) or die("Cannot write cache config file");
  print F "center=".(($topLeftCoords[0]+$bottomRightCoords[0])/2).",".(($topLeftCoords[1]+$bottomRightCoords[1])/2).",".$zoomLevel.",".$MapType."\n";
  close(F);
}

# tv-arnd begin
  print "\nEnter short Map Name Prefix to individualize map (optional): ";
  chomp ($MapNamePrefix = <>); # if this is nothing thats fine
  if ($MapNamePrefix) {$MapNamePrefix .= "_";}
# tv-arnd end


print <<EOF;
------------------------
Ready to cache $numTiles tiles in first Zoom-Level...

Type: $MapType
Zoom: $zoomLevel

topLeftTileX:     $topLeftTileX      topLeftTileY: $topLeftTileY
bottomRightTileX: $bottomRightTileX  bottomRightTileY: $bottomRightTileY
ZoomLevels:       $ZoomLevels_wanted

Press enter to continue.
EOF

$_ = <>;

# -----------now all input data is complete----------------------

# tv-arnd begin

@ZoomLevelLines; # global array to store CompeGPS multi-zoom definitions, one line per zoom-level in imp-file
                 # $nlevel_counter this starts from number of total zoom-levels-1, used in multi-level imp-file
                 # $nlevel_counter will be decreased for each zoom-level, down to 0
my $actualTileCount = 0; # used for sleeping?

for (my $nlevel_counter = $ZoomLevels_wanted-1; $nlevel_counter >= 0; $nlevel_counter--) { # should not go below 0
  $numTiles = (abs($topLeftTileX-$bottomRightTileX)+1)*(abs($topLeftTileY-$bottomRightTileY)+1);
  $numTiles *= (index($MapType,"Hyb") != -1 && ($MapType ne "MicrosoftHyb")) ? 2 : 1;

  print "Zoom-Level: $zoomLevel\n"; print "\$numTiles: $numTiles\n";

  print "*** Caching Started ***\n\n";
    &downloadTiles($MapType, $topLeftTileX, $topLeftTileY, $bottomRightTileX, $bottomRightTileY, 1);               print "\n*** Caching Complete ***\n";
  print "\n*** Creating CompeGPS-Map ***\n";
    &montageTiles_rectangle ($topLeftTileX, $topLeftTileY, $bottomRightTileX, $bottomRightTileY, $nlevel_counter); print "*** Creating CompeGPS-Map Complete ***\n\n";

  # double the resolution , i.e. zoom into next higher zoomlevel for next iteration
  $zoomLevel++; 
  $topLeftTileX *= 2; $topLeftTileY *= 2;
  $bottomRightTileX = $bottomRightTileX * 2 + 1; $bottomRightTileY = $bottomRightTileY * 2 +1;

}
# finally call this with the parameters of map with highest zoom-level (map/images with maximum dimensions)
# need to devide by 2 to reverse last doubling of for-loop above
# call only if there are more than 1 zoom-levels
if ($ZoomLevels_wanted > 1) {
  $zoomLevel--; 
  &makeCalFile_multiZoom ($topLeftTileX/2, $topLeftTileY/2, ($bottomRightTileX-1)/2, ($bottomRightTileY-1)/2, $MapNamePrefix);
}
print "\n*** Complete ***\n";
# tv-arnd end

exit;

#MapTileCacher Subroutines

sub convertCoordindates2Tiles
{
    my $lat = shift;
    my $long = shift;
    my $z = shift;

    my $pixelsPerTile = 256;

    my $mapsize = 2**($z+(log($pixelsPerTile)/log(2)));
    my $origin = $mapsize / 2;

    my $longdeg = abs(-180 - $long);
    my $longppd = $mapsize / 360;
    my $longppdrad = $mapsize/(2*pi);
    my $pixelx = $longdeg * $longppd;
    my $longtiles = $pixelx  / $pixelsPerTile;

    my $tilex = cast2Integer($longtiles);

    my $e = sin($lat*(1/180.*pi));
    if ($e>0.9999) { $e=0.9999; }
    if ($e<-0.9999) { $e=-0.9999; }

    my $pixely = $origin + 0.5*log((1+$e)/(1-$e)) * (-$longppdrad);
    my $tiley = cast2Integer($pixely / $pixelsPerTile);

    return ($tilex, $tiley);
}

sub downloadTiles
{
  my $MapType = shift;
  my $topLeftTileX = shift;
  my $topLeftTileY = shift;
  my $bottomRightTileX = shift;
  my $bottomRightTileY = shift;
  my $rec = shift;
  
  my $tileCount = 0;

  if(index($MapType,"Hyb") != -1 && ($MapType ne "MicrosoftHyb") && ($rec == 1))
  {
      my $MapSource = substr($MapType,0,index($MapType,"Hyb"));
      if (index($MapType,"Ter") != -1)
      {
        downloadTiles($MapSource, $topLeftTileX, $topLeftTileY, $bottomRightTileX, $bottomRightTileY, 0);
      }
      else
      {
        downloadTiles($MapSource . "Sat", $topLeftTileX, $topLeftTileY, $bottomRightTileX, $bottomRightTileY, 0);
      }
      downloadTiles($MapSource . "Hyb", $topLeftTileX, $topLeftTileY, $bottomRightTileX, $bottomRightTileY, 0);
  }
  else
  {
    for ($y = $topLeftTileY; $y <= $bottomRightTileY; $y++)
    {
      for ($x = $topLeftTileX; $x <= $bottomRightTileX; $x++)
      {
        $tileCount++;

        # avoid sleeping for #0
        if ((($actualTileCount+1) % $sleepEvery) == 0 && index($MapType,"Sat") != -1)
        {
          print "* Sleeping $sleepTime seconds\n";
          sleep($sleepTime);
        }

        my $localFileName = $dir_TopLevel . $systemDirDivider . $MapType . "_" .$zoomLevel;
        if ($hash_size > 1) {
          $localFileName .= $systemDirDivider . (($x*256+$y)%$hash_size) . $systemDirDivider . "$x\_$y";
        }
        else {
          $localFileName .= $systemDirDivider . int($x/$tpfx) . "_" . int($y/$tpfy);
        }
        $localFileName .= ".mgm";
        $dx = $x % $tpfx;
        $dy = $y % $tpfy;

        # Cache Source
        my $localFileName_src = $dir_TopLevel_src . $systemDirDivider . $MapType . "_" .$zoomLevel;
        if ($hash_size_src > 1) {
          $localFileName_src .= $systemDirDivider . (($x*256+$y)%$hash_size_src) . $systemDirDivider . "$x\_$y";
        }
        else {
          $localFileName_src .= $systemDirDivider . int($x/$tpfx_src) . "_" . int($y/$tpfy_src);
        }
        $localFileName_src .= ".mgm";
        $dx_src = $x % $tpfx_src;
        $dy_src = $y % $tpfy_src;

        my $url = getTileURL($MapType,$x,$y);

        $found = checkCached($localFileName,$dx,$dy);

        if($found == 0)
        {
          if($cache_source == 1)
          {
            print "* Reading tile $tileCount of $numTiles from cached file $localFileName_src\n";
            $found = getCached($localFileName_src,$dx_src,$dy_src);
          }
          if($found == 0) {
            my $referer = (index($MapType,"Google") != -1) ? $systemRefererParam : "";
            $wgetFileCommand = $wget . $referer. "\"" . $url . "\"" ;
            print "* Downloading tile $tileCount of $numTiles from $url\n";
            if (index($MapType,"Sat") != -1) {$actualTileCount++;}
            system($wgetFileCommand);
          }
# by tv-arnd, copy the temp-tiles, numbered with x_y in diectory $dir_TypeZoomLevel
          copyDownloadedTile ("TempTile.$pid",$localFileName,$x,$y);
# by tv-arnd
          cacheDownloadedTile("TempTile.$pid",$localFileName,$dx,$dy);
        }

        else {print "* Skipping tile $tileCount of $numTiles from $url (already cached).\n";}
      }
    }
  }
}

# check if the map tile is already cached in $localFilename as ($dx,$dy)
sub checkCached
{
  my $localFileName = shift;
  my $dx = shift;
  my $dy = shift;
  local $buf = '';

  if (-f $localFileName == FALSE) {
    return 0;
  }

  # if the file exists and we have 1 tile/file, don't verify anything else
  elsif ($tiles_per_file == 1) {
    return 1;
  }

  sysopen(FD, $localFileName, O_RDONLY);
  binmode(FD);
  sysread(FD, $buf, 6*$tiles_per_file+2);
  $num = (ord(substr($buf,0,1))<< 8 ) + ord(substr($buf,1,1));
  for (my $i=0; $i<$num; $i++) {
    if (ord(substr($buf,2+$i*6,1)) == $dx && ord(substr($buf,2+$i*6+1,1)) == $dy) {
      close(FD);
      return 1;
    }
  }
  close(FD);
  return 0;
}

# get the map tile already cached in $localFilename as ($dx,$dy)
sub getCached
{
  my $localFileName = shift;
  my $dx = shift;
  my $dy = shift;
  local $buf = '';

  my $ofs_prev = 0;
  my $ofs_next = 2+6*$tiles_per_file_src;

  # print "   getCached($localFileName $dx $dy)\n";

  if (-f $localFileName == FALSE) {
    return 0;
  }

  # if the file exists and we have 1 tile/file, just copy the source file
  elsif ($tiles_per_file_src == 1) {
    $copyCommand = "$systemCopyCommand $localFileName TempTile.$pid";
    # print "  executing: $copyCommand";
    system($copyCommand);
    return 1;
  }

  sysopen(FD, $localFileName, O_RDONLY);
  binmode(FD);
  sysread(FD, $buf, 6*$tiles_per_file_src+2);
  $num = (ord(substr($buf,0,1))<<8)+ord(substr($buf,1,1));
  for (my $i=0; $i<$num; $i++) {
    $ofs_prev = $ofs_next;
    $ofs_next = (ord(substr($buf,2+$i*6+2,1))<<24)+
              (ord(substr($buf,2+$i*6+3,1))<<16)+
              (ord(substr($buf,2+$i*6+4,1))<<8)+
               ord(substr($buf,2+$i*6+5,1));
    if (ord(substr($buf,2+$i*6,1)) == $dx && ord(substr($buf,2+$i*6+1,1)) == $dy) {
    # read image at proper offset and write to TempTile.$pid
    sysseek(FD, $ofs_prev, 0);
    sysread(FD, $buf, $ofs_next - $ofs_prev);
      close(FD);
      sysopen(FD, "TempTile.$pid", O_WRONLY|O_CREAT);
    binmode(FD);
    syswrite(FD, $buf, $ofs_next - $ofs_prev);
    close(FD);
      return 1;
    }
  }
  close(FD);
  return 0;
}

sub cacheDownloadedTile
{
  my $tempTile = shift;
  my $localFileName = shift;
  my $dx = shift;
  my $dy = shift;
  local $buf = '';

  if(-s $tempTile)
  {
    my $dir_TopLevel;
    my $dir_TypeZoomLevel;
    my $dir_HashLevel;
    if ($hash_size > 1) {
      ($dir_TopLevel, $dir_TypeZoomLevel, $dir_HashLevel)
      = split(($systemDirDivider eq "\\") ? "$systemDirDivider$systemDirDivider" : "$systemDirDivider",$localFileName);
    }

    else {
      ($dir_TopLevel, $dir_TypeZoomLevel)
      = split(($systemDirDivider eq "\\") ? "$systemDirDivider$systemDirDivider" : "$systemDirDivider",$localFileName);
    }

    $dir_TypeZoomLevel = $dir_TopLevel . $systemDirDivider . $dir_TypeZoomLevel;
    if(-d $dir_TypeZoomLevel == FALSE){system("mkdir $dir_TypeZoomLevel");}

    if ($tiles_per_file == 1) {
      if ($hash_size > 1) {
        $dir_HashLevel = $dir_TypeZoomLevel . $systemDirDivider . $dir_HashLevel;
        if(-d $dir_HashLevel == FALSE){system("mkdir $dir_HashLevel");}
      }
      system("$systemMoveCommand $tempTile $localFileName");
      print "\tTile cached at $localFileName\n";
      return;
    }

    # append to the file, write info and offset to header
    my $num = 0;
    my $ofs = $tiles_per_file*6+2;
    $existed = ((-f $localFileName != FALSE) ? 1 : 0);
    sysopen($fh, $localFileName, O_RDWR|O_CREAT);
    binmode($fh);
    if ($existed == 1) {
      sysread($fh, $buf, $ofs);
      $num = (ord(substr($buf,0,1))<<8)+ord(substr($buf,1,1));
      $ofs = (ord(substr($buf,2+$num*6-6+2,1))<<24)+
             (ord(substr($buf,2+$num*6-6+3,1))<<16)+
             (ord(substr($buf,2+$num*6-6+4,1))<<8)+
             ord(substr($buf,2+$num*6-6+5,1));
    }
    else {
      $buf = chr(0) x $ofs;
      syswrite($fh, $buf, $ofs);
    }
    my $totalsize = 0;
    sysseek($fh, $ofs, 0);
    sysopen(F2, $tempTile, O_RDONLY);
    binmode(F2);
    while ((my $read = sysread(F2, $buf, 4096)) > 0) {
      $totalsize += $read;
      syswrite($fh, $buf, $read);
    }
    close(F2);
    # new offset
    $totalsize += $ofs;

    # update header - number of tiles
    $buf = "" . chr(($num+1)>>8) . chr(($num+1)&255);
    sysseek($fh, 0, 0);
    syswrite($fh, $buf, 2);

    # update header - new tile data
    $buf = "" . chr($dx) . chr($dy) . chr($totalsize>>24) . chr(($totalsize>>16)&255) . chr(($totalsize>>8)&255) . chr($totalsize&255);
    sysseek($fh, 6*$num+2, 0);
    syswrite($fh, $buf, 6);
    close($fh);

    # delete temp tile
    system("$systemDeleteCommand $tempTile");


    print "\tTile cached at $localFileName\n"
  }
  else {system("$systemDeleteCommand $tempTile"); print "\tERROR: Requested tile did not exist.\n"}
}

sub getTileURL
{
  my $MapType = shift;
  my $x = shift;
  my $y = shift;

  my $url = "";

  if($MapType eq "GoogleMap" || $MapType eq "GoogleHyb" || $MapType eq "GoogleTer")
  {
    $url = "http://mt" . (($x+$y)&3) . ".google.com/".(($MapType eq "GoogleTer") ? "m" : "v")."t/v=";
    $url .= ($MapType eq "GoogleMap") ? "w2.104" :
            ($MapType eq "GoogleHyb") ? "w2t.104" : "w2p.87";
    $url .= "&hl=en&x=$x&y=$y&z=$zoomLevel";
  }
  elsif($MapType eq "GoogleChina")
  {
    $url = "http://mt" . (($x+$y)&3) . ".google.cn/vt/v=w2.104&hl=en&gl=cn&x=$x&y=$y&z=$zoomLevel";
  }
  elsif($MapType eq "GoogleSat")
  {
    if ($use_proxy)
    {
      $url = $proxy_list[int(rand(@proxy_list))] . "http/khm";
    }
    else
    {
      $url = "http://khm";
    }

    $url .= (($x+$y)&3) . ".google.com/kh/v=43&hl=en&x=$x&y=$y&z=$zoomLevel";
  }
  elsif(index($MapType,"YahooIn") != -1)
  {
    $url = ($MapType eq "YahooInMap") ? "http://maps.yimg.com/hw/tile?locale=en&imgtype=png&yimgv=1.2&v=4.1" :
           "http://maps.yimg.com/hw/tile?imgtype=png&yimgv=0.95&t=h";
    $url .= "&x=$x&y=" . (((1 << $zoomLevel) >> 1)-1-$y) . "&z=" . (17-$zoomLevel+1);
  }
  elsif(index($MapType,"Yahoo") != -1)
  {
    $url = "http://maps".(1+(($x+$y)%3)).".yimg.com/";
    $url .= ($MapType eq "YahooMap") ? "tl?v=4.2" :
       ($MapType eq "YahooSat") ? "ae/ximg?v=1.9&t=a&s=256" :
       "hx/tl?v=4.2&t=h";
    $url .= "&.intl=en&x=$x&y=" . (((1 << $zoomLevel) >> 1)-1-$y) . "&z=" . ($zoomLevel+1) . "&r=1";
  }
  elsif(index($MapType,"Microsoft") != -1)
  {
    $digit = ((($y & 1) << 1) + ($x & 1));
    if ($MapType eq "MicrosoftBrMap") {
      $digit++;
      $url = "http://imakm" . $digit . ".maplink3.com.br/maps.ashx?v=";
    }
    else
    {
      $url  = ($MapType eq "MicrosoftHyb") ?  "http://h" : ($MapType eq "MicrosoftSat") ?  "http://a" : "http://r";
      $url .= $digit . ".ortho.tiles.virtualearth.net/tiles/";
      $url .= ($MapType eq "MicrosoftHyb") ?  "h" : ($MapType eq "MicrosoftSat") ?  "a" : "r";
    }
    for($i=$zoomLevel-1; $i >= 0; $i--)
    {
              $url = $url . ((((($y >> $i) & 1) << 1) + (($x >> $i) & 1)));
    }
    if ($MapType eq "MicrosoftBrMap")
    {
      $url .= "|t&call=2.2.4";
    }
    else
    {
      $url .= ($MapType eq "MicrosoftMap" || $MapType eq "MicrosoftTer") ?  ".png" : ".jpeg";
      $url .= "?g=321&mkt=en-us";
      if ($MapType eq "MicrosoftTer")
      {
        $url .= "&shading=hill";
      }
    }
  }
  elsif(index($MapType,"OpenStreetMap") != -1)
  {
    $url  = "http://tile.openstreetmap.org/" . $zoomLevel . "/" . $x . "/" . $y . ".png";
  }
  elsif(index($MapType,"Osmarender") != -1)
  {
    $url = "http://tah.openstreetmap.org/Tiles/tile/" . $zoomLevel . "/" . $x . "/" . $y . ".png";
  }
  elsif(index($MapType,"OpenAerialMap") != -1)
  {
    $url = "http://tile.openaerialmap.org/tiles/?v=mgm&layer=openaerialmap-900913&x=$x&y=$y&zoom=" . (17-$zoomLevel);
  }
  elsif(index($MapType,"OpenCycleMap") != -1)
  {
    $url = "http://andy.sandbox.cloudmade.com/tiles/cycle/" . $zoomLevel . "/" . $x . "/" . $y . ".png";
  }

  return $url;
}

#Utility Subroutines

sub cast2Integer
{
    my $f = shift;
    if ($f < 0) {return ceil($f);}
    else {return floor($f);}
}

# tv-arnd's Subroutines

# conversion: lon/lat to tile numbers
# from http://wiki.openstreetmap.org/wiki/Slippy_map_tilenames
sub getTileNumber {
  my ($lat,$lon,$zoom) = @_;
  my $xtile = int( ($lon+180)/360 *2**$zoom ) ;
  my $ytile = int( (1 - log(tan($lat*pi/180) + sec($lat*pi/180))/pi)/2 *2**$zoom ) ;
  return(($xtile, $ytile));
}


# conversion: tile numbers to lon/lat 
# from http://wiki.openstreetmap.org/wiki/Slippy_map_tilenames
sub Project {
  my ($X,$Y, $Zoom) = @_;
  my $Unit = 1 / (2 ** $Zoom);
  my $relY1 = $Y * $Unit;
  my $relY2 = $relY1 + $Unit;

  # note: $LimitY = ProjectF(degrees(atan(sinh(pi)))) = log(sinh(pi)+cosh(pi)) = pi
  # note: degrees(atan(sinh(pi))) = 85.051128..
  #my $LimitY = ProjectF(85.0511);

  # so stay simple and more accurate
  my $LimitY = pi;
  my $RangeY = 2 * $LimitY;
  $relY1 = $LimitY - $RangeY * $relY1;
  $relY2 = $LimitY - $RangeY * $relY2;
  my $Lat1 = ProjectMercToLat($relY1);
  my $Lat2 = ProjectMercToLat($relY2);
  $Unit = 360 / (2 ** $Zoom);
  my $Long1 = -180 + $X * $Unit;
  return(($Lat2, $Long1, $Lat1, $Long1 + $Unit)); # S,W,N,E
 }
 sub ProjectMercToLat($){
  my $MercY = shift();
  return( 180/pi* atan(sinh($MercY)));
 }
 sub ProjectF
 {
  my $Lat = shift;
  $Lat = deg2rad($Lat);
  my $Y = log(tan($Lat) + (1/cos($Lat)));
  return($Y);
 }

# save the temp-tile as raw bitmap
sub copyDownloadedTile {
  my $tempTile = shift;
  my $localFileName = shift;
  my $x = shift;
  my $y = shift;
  my $dir_TopLevel;
  my $dir_TypeZoomLevel;
  my $dir_HashLevel;
  if ($hash_size > 1) {
    ($dir_TopLevel, $dir_TypeZoomLevel, $dir_HashLevel)
    = split(($systemDirDivider eq "\\") ? "$systemDirDivider$systemDirDivider" : "$systemDirDivider",$localFileName);
  }

  else {
    ($dir_TopLevel, $dir_TypeZoomLevel)
    = split(($systemDirDivider eq "\\") ? "$systemDirDivider$systemDirDivider" : "$systemDirDivider",$localFileName);
  }

  $dir_TypeZoomLevel = $dir_TopLevel . $systemDirDivider . $dir_TypeZoomLevel;
  if(-d $dir_TypeZoomLevel == FALSE){system("mkdir $dir_TypeZoomLevel");}
  
  system("$systemCopyCommand $tempTile $dir_TypeZoomLevel$systemDirDivider" . "Tile_$x\_$y.png");
  print "\tTile copied at $systemCopyCommand $tempTile $dir_TypeZoomLevel$systemDirDivider" . "Tile_$x\_$y.png" . "\n";
}

# combine raw image tiles into combined image
# combined image may not be square !!!
sub montageTiles_rectangle
{
  my $topLeftTileX = shift;
  my $topLeftTileY = shift;
  my $bottomRightTileX = shift;
  my $bottomRightTileY = shift;
  my $nlevel = shift; # multizoom imp file nlevel number

  my $dir_TypeZoomLevel = $dir_TopLevel . $systemDirDivider . $MapType . "_" .$zoomLevel;
  my $MapFileName; # will be someting like $MapNamePrefix_OpenStreetMap_Z15_0xx_0yy without file extension !
  #print "DIR " . $dir_TypeZoomLevel . "\n"; 
  chdir($dir_TypeZoomLevel) or die "Cant chdir to $dir_TypeZoomLevel $!"; # go to zoomlevel dir to shorten montage-call - do not forget to change back!

  # build a string containing file names of all tiles, but number of tiles is limited to a max. of  $tiles_per_file


  my $diffX = $bottomRightTileX - $topLeftTileX +1; # here 1 more because 0 counts
  my $diffY = $bottomRightTileY - $topLeftTileY +1;
  #print "Diff X / Y: " . $diffX . " / " . $diffY . "\n"; print "Number of Tiles: " . $diffX * $diffY . "\n";
  
  my $tiles_per_file_X = int (sqrt ($tiles_per_file)); 
  my $tiles_per_file_Y = int (sqrt ($tiles_per_file)); # Y seems to be not required, i.e. equals width as we want squares
  #print "Wanted Tiles per File (must be 2^n): $tiles_per_file\t=> n: $tiles_per_file_X\n";
  
  my $Map_Index_max_digits = int (log ( ( ($diffX <= $diffY) ? $diffY : $diffX)/$tiles_per_file_X ) / log(10)) +1+1; # maximum number of digits for map XY-index, for later creating formatted file names, plus one more for leading zero
  
  # loop to arrange a maximum number of tiles in squares, if not possible make rectangles (at the bottom end, right end)
  my $tile_counter = 1;                       # overall tile counter
  my $tile_counterX; my $tile_counterY;       # tile counter within inner loop to get option-string for ImageMagick montage e.g. "4x4"
  my $map_counterX = 0; my $map_counterY = 0; # tile counter within outer loop to get indices to position maps in IMP-file
  my $montage_tile;                           # option-string for ImageMagick montage e.g. "4x4"
  my $tile_files_list ="";                    # holds the file names of the files to be montage'd
  my $pixelsPerTile = 256;

  # outer loop to define starting positions for inner loop
  for (my $startTileX = $topLeftTileX; $startTileX <= $bottomRightTileX; $startTileX += $tiles_per_file_X) {
    for (my $startTileY = $topLeftTileY; $startTileY <= $bottomRightTileY; $startTileY += $tiles_per_file_X) {

      # inner xy-loop to generate square maps of a given width (if possible) at starting position defined by the outer loop
      $tile_counterY = 0; $tile_files_list ="";
      for (my $y = $startTileY; ($y < ($startTileY + $tiles_per_file_X )) && ($y <= $bottomRightTileY); $y++ ) {
	
	$tile_counterX = 0;
	for (my $x = $startTileX; ($x < ($startTileX + $tiles_per_file_X)) && ($x <= $bottomRightTileX) ; $x++ ) {
	  $tile_files_list .= "Tile_$x\_$y.png ";  
	  #print "$tile_counter. \tLoop-Index: x=$x, y=$y\n";
	  $tile_counter++; $tile_counterX++;
	}
	$tile_counterY++;
      }
      $montage_tile = $tile_counterX . "x". $tile_counterY; # option-string for ImageMagick montage e.g. "4x4  
      $MapFileName = $MapNamePrefix . $MapType . "_Z" . $zoomLevel . "_" . (sprintf "%0$Map_Index_max_digits.0f", $map_counterX) . "_" . (sprintf "%0$Map_Index_max_digits.0f", $map_counterY);
      printf "Writing $MapFileName.png file with $montage_tile tiles and IMP-Map Index: %0$Map_Index_max_digits.0f %0$Map_Index_max_digits.0f\n", $map_counterX, $map_counterY; # here execute montage command
      #print "tile_file_list:\n";

      system($IMmontage . " -monitor -units Undefined -density 0 -depth 8 -geometry +0+0 -tile $montage_tile " . $tile_files_list . $MapFileName . ".png"); # DPI's are undefined, CompeGPS requires color max. depth = 8
      makeCalFile_singleZoom ($startTileX, $startTileY, $startTileX+$tile_counterX-1, $startTileY+$tile_counterY-1, $MapFileName);

      #print "tile_counterX and tile_counterX: $tile_counterX, $tile_counterY\n";
      $map_counterY++;
    }
    $map_counterY = 0; $map_counterX++;
  }
    chdir(".." . $systemDirDivider . "..") or die "Cant chdir to " . ".." . $systemDirDivider . ".." . "$!"; # 2 dir-levels up, not sure if this works for MS Win
    #print "back in original dir: "; system("pwd");
   
   # writing specific line for multi-level zoom imp-file
   # need the pixels of first map saved (upper left map - this map becomes a tile within the multi-level zoom imp file)
   # this number can be the maximum number of pixels (upper value as defind by $tiles_per_file) or less if diffXY is smaller than $tiles_per_file_XY
   my $pixelsPerMapX = (($diffX <= $tiles_per_file_X) ? $diffX : $tiles_per_file_X) * $pixelsPerTile;
   my $pixelsPerMapY = (($diffY <= $tiles_per_file_Y) ? $diffY : $tiles_per_file_Y) * $pixelsPerTile; 
   $MapFileName = $MapNamePrefix . $MapType . "_Z" . $zoomLevel;
   push (@ZoomLevelLines, "<level nlevel=\"$nlevel\" filename=\"$MapFileName\_%0" . $Map_Index_max_digits . "d_%0" . $Map_Index_max_digits . "d.RTMAP\" filetilex=\"$pixelsPerMapX\" filetiley=\"$pixelsPerMapY\" tilex=\"256\" tiley=\"256\" txyorder=\"0\" tx0=\"0\" ty0=\"0\" dtx=\"1\" dty=\"1\" yupfiletiles=\"0\"/>");
}

# create and write simple IMP (CompeGPS map calibration) file
# only one zoom-level, i.e. no zoom level at all
sub makeCalFile_singleZoom
{
  my $topLeftTileX = shift;
  my $topLeftTileY = shift;
  my $bottomRightTileX = shift;
  my $bottomRightTileY = shift;
  my $MapFileName = shift;

  my $CalFileName = $MapFileName . ".imp"; # CompeGPS imp file
  $MapFileName = $MapFileName . ".png";

  my $diffX = $bottomRightTileX - $topLeftTileX +1; # here 1 more than real
  my $diffY = $bottomRightTileY - $topLeftTileY +1;

  my $pixelsPerTile = 256;
  my $pixelsPerMapX = $diffX * $pixelsPerTile;
  my $pixelsPerMapY = $diffY * $pixelsPerTile;

  (my $ul_tile_lat2, my $ul_tile_long1, my $ul_tile_lat1, my $ul_tile_long2) = Project ($topLeftTileX, $topLeftTileY, $zoomLevel); # upper left tile
  #print "ul_tile_lat2:\t" . $ul_tile_lat2 . "\tul_tile_long1:\t" . $ul_tile_long1 . "\tul_tile_lat1:\t" . $ul_tile_lat1 . "\tul_tile_long2:\t" . $ul_tile_long2 . "\n";
  (my $br_tile_lat2, my $br_tile_long1, my $br_tile_lat1, my $br_tile_long2) = Project ($bottomRightTileX, $bottomRightTileY, $zoomLevel); # bottom right tile
  #print "br_tile_lat2:\t" . $br_tile_lat2 . "\tbr_tile_long1:\t" . $br_tile_long1 . "\tbr_tile_lat1:\t" . $br_tile_lat1 . "\tbr_tile_long2:\t" . $br_tile_long2 . "\n";

  
  print "Writing Calibration File: $CalFileName\n\n";
  open(F, ">".$CalFileName) or die("Cannot write calibration file");

  print F <<EOF;
CompeGPS MAP File
<Header>
Version=2
VerCompeGPS=6.8.5
Projection=2,Mercator,
Coordinates=1
Datum=WGS 84
</Header>
<Map>
Bitmap=$MapFileName
BitsPerPixel=0
BitmapWidth=$pixelsPerMapX
BitmapHeight=$pixelsPerMapY
Type=10
</Map>
<Calibration>
P0=0.00000000,0.00000000,A,$ul_tile_long1,$ul_tile_lat1
P1=$pixelsPerMapX,$pixelsPerMapY,A,$br_tile_long2,$br_tile_lat2
P2=0.00000000,$pixelsPerMapY,A,$ul_tile_long1,$br_tile_lat2
P3=$pixelsPerMapX,0.00000000,A,$br_tile_long2,$ul_tile_lat1
</Calibration>
<MainPolygonBitmap>
M0=0.0,0.0
M1=$pixelsPerMapX,0.0
M2=$pixelsPerMapX,$pixelsPerMapY
M3=0.0,$pixelsPerMapY
</MainPolygonBitmap>
EOF
  close(F);
}

# create and write multi-zoom IMP (CompeGPS map calibration) file
sub makeCalFile_multiZoom
{
  my $topLeftTileX = shift;
  my $topLeftTileY = shift;
  my $bottomRightTileX = shift;
  my $bottomRightTileY = shift;
  my $MapFileName = shift;

  my $CalFileName =  $dir_TopLevel . $systemDirDivider . $MapFileName . $MapType . "ML.imp"; # ML - Multilevel CompeGPS imp file

  my $diffX = $bottomRightTileX - $topLeftTileX +1; # here 1 more than real, i.e. start of next tile = end of last tile - is this ok?
  my $diffY = $bottomRightTileY - $topLeftTileY +1; # same, same

  my $pixelsPerTile = 256;
  my $pixelsPerMapX = $diffX * $pixelsPerTile;
  my $pixelsPerMapY = $diffY * $pixelsPerTile;

  (my $ul_tile_lat2, my $ul_tile_long1, my $ul_tile_lat1, my $ul_tile_long2) = Project ($topLeftTileX, $topLeftTileY, $zoomLevel); # upper left tile
  #print "ul_tile_lat2:\t" . $ul_tile_lat2 . "\tul_tile_long1:\t" . $ul_tile_long1 . "\tul_tile_lat1:\t" . $ul_tile_lat1 . "\tul_tile_long2:\t" . $ul_tile_long2 . "\n";
  (my $br_tile_lat2, my $br_tile_long1, my $br_tile_lat1, my $br_tile_long2) = Project ($bottomRightTileX, $bottomRightTileY, $zoomLevel); # bottom right tile
  #print "br_tile_lat2:\t" . $br_tile_lat2 . "\tbr_tile_long1:\t" . $br_tile_long1 . "\tbr_tile_lat1:\t" . $br_tile_lat1 . "\tbr_tile_long2:\t" . $br_tile_long2 . "\n";

  @ZoomLevelLines = reverse @ZoomLevelLines;
  my $ZoomLevelLines_formatted;
  foreach (@ZoomLevelLines) {
    $ZoomLevelLines_formatted = $ZoomLevelLines_formatted . $_ . "\n";
  }
 
  print "Writing Multi-Level Zoom Calibration File: $CalFileName\n";
  open(F, ">".$CalFileName) or die("Cannot write calibration file");

  print F <<EOF;
CompeGPS MAP File
<Header>
Version=2
VerCompeGPS=6.8.5
Projection=0,Mercator,
Coordinates=1
Datum=WGS 84
</Header>
<Map>
Bitmap=nothing here
BitsPerPixel=0
BitmapWidth=$pixelsPerMapX
BitmapHeight=$pixelsPerMapY
Type=10
<BitmapData>
<rtile>
<CompeGPSRasterTiled pixelstotx="$pixelsPerMapX" pixelstoty="-$pixelsPerMapY">
$ZoomLevelLines_formatted
</CompeGPSRasterTiled>
</rtile>
</BitmapData>
</Map>
<Calibration>
P0=0.00000000,0.00000000,A,$ul_tile_long1,$ul_tile_lat1
P1=$pixelsPerMapX,$pixelsPerMapY,A,$br_tile_long2,$br_tile_lat2
P2=0.00000000,$pixelsPerMapY,A,$ul_tile_long1,$br_tile_lat2
P3=$pixelsPerMapX,0.00000000,A,$br_tile_long2,$ul_tile_lat1
</Calibration>

EOF
  close(F);
}